/*
 *  tags.c
 *  Routines for reading and modifying id3 tags, using libid3tag.
 *
 *  Created by Alexander Thomas on 22/07/06.
 * TODO: convert to a C++ class that hides all the loathsome ugliness
 *       of the C code.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "tags.h"


/* type must be a 4-byte identifier like TALB, TXXX, ...
 * field2 may be NULL if this type has only 1 field. */
struct id3_frame * makeID3Frame( const char *type, const char *field1,
                                 const char *field2, const char *field3 )
{
	struct id3_frame *newFrame = id3_frame_new(type);
	
#ifdef DEEPDEBUG
	fprintf( stderr, "Created %s frame with %u fields\n", type, newFrame->nfields );
	for( int i=0; i<newFrame->nfields; i++ ) {
		fprintf( stderr, "Field %d type: %u\n", i, id3_field_type(&newFrame->fields[i]) );
	}
	fprintf( stderr, "Text encoding: %d\n", id3_field_gettextencoding( &newFrame->fields[0] ) );
#endif
	
	setField( &newFrame->fields[1], field1 );
	if( newFrame->nfields > 2 )
		setField( &newFrame->fields[2], field2 );
	if( newFrame->nfields > 3 )
		setField( &newFrame->fields[3], field3 );
	if( newFrame->nfields > 4 )
		fprintf( stderr, "ERROR: makeID3Frame not implemented for %d fields!\n", newFrame->nfields );
	
	return newFrame;
}


struct id3_frame * makeID3BinaryFrame( const char *type, const id3_byte_t *data, id3_length_t length )
{
	struct id3_frame *newFrame = id3_frame_new(type);
	
#ifdef DEEPDEBUG
	fprintf( stderr, "Created %s frame with %u fields\n", type, newFrame->nfields );
	fprintf( stderr, "Field 0 type: %u\n", id3_field_type( &newFrame->fields[0]) );
#endif
	
	id3_field_setbinarydata( &newFrame->fields[0], data, length );
	
	return newFrame;
}


int setField( union id3_field *field, const char *contents )
{
	int retVal;
	enum id3_field_type fieldType = id3_field_type(field);
	if( fieldType == ID3_FIELD_TYPE_LATIN1 )
		retVal = id3_field_setlatin1( field, (id3_latin1_t*)contents );
	else if( fieldType == ID3_FIELD_TYPE_LATIN1FULL )
		retVal = id3_field_setfulllatin1( field, (id3_latin1_t*)contents );
	else if( fieldType == ID3_FIELD_TYPE_LANGUAGE )
		retVal = id3_field_setlanguage( field, contents );
	else if( fieldType == ID3_FIELD_TYPE_STRING ||
	         fieldType == ID3_FIELD_TYPE_STRINGFULL ) {
		/* Note: The need for 4-byte character strings stems from
		 * the fact that UTF-8 characters can be up to 4 bytes long. */
		id3_ucs4_t *ucs4String = id3_latin1_ucs4duplicate( (id3_latin1_t*)contents );
		if( fieldType == ID3_FIELD_TYPE_STRING )
			retVal = id3_field_setstring( field, ucs4String );
		else
 			retVal = id3_field_setfullstring( field, ucs4String );
		free(ucs4String);
	}
	else {
		retVal = -1;
		fprintf( stderr, "ERROR: setField not implemented for field type %d!\n", fieldType );
	}
	return retVal;
}


unsigned char * getField( union id3_field *field )
{
	unsigned char *output;
	enum id3_field_type fieldType = id3_field_type(field);
	if( fieldType == ID3_FIELD_TYPE_LATIN1 || fieldType == ID3_FIELD_TYPE_LANGUAGE )
		output = dupString( id3_field_getlatin1(field) );
	else if( fieldType == ID3_FIELD_TYPE_LATIN1FULL )
		output = dupString( id3_field_getfulllatin1(field) );
	else if( fieldType == ID3_FIELD_TYPE_STRING ) {
		const id3_ucs4_t *ucs4String = id3_field_getstring(field);
		output = id3_ucs4_latin1duplicate(ucs4String);
	}
	else if( fieldType == ID3_FIELD_TYPE_STRINGFULL ) {
		const id3_ucs4_t *ucs4String = id3_field_getfullstring(field);
		output = id3_ucs4_latin1duplicate(ucs4String);
	}
	else {
		output = NULL;
		fprintf( stderr, "ERROR: getField not implemented for field type %d!\n", fieldType );
	}
	return output;
}


inline int id3_ucs4cmp( const id3_ucs4_t *ucs4String, const char *string )
{
	char *tmp = (char*)id3_ucs4_latin1duplicate(ucs4String);
	int ret = strcmp( tmp, string );
	free(tmp);
	return ret;
}


struct id3_frame * findFrameWithLabel( struct id3_tag *fileTag, const char *type,
                                       const char *label, int labelPos )
{
	struct id3_frame *frame;
	unsigned int i = 0;
	
	frame = id3_tag_findframe( fileTag, type, 0 );
	while( frame != NULL ) {
		const id3_ucs4_t *fieldName = id3_field_getstring( id3_frame_field(frame, labelPos) );
		if( id3_ucs4cmp( fieldName, label ) == 0 )
			return frame;
		
		frame = id3_tag_findframe( fileTag, type, ++i );
	}
	
	return NULL;
}


struct id3_frame * findTXXX( struct id3_tag *fileTag, const char *label )
{
	return findFrameWithLabel( fileTag, "TXXX", label, 1 );
}


int setTXXX( struct id3_tag *fileTag, const char *label, const char *contents )
{
	struct id3_frame *tmpFrame;
	
	tmpFrame = findTXXX( fileTag, label );
	if( tmpFrame != NULL ) {
		setField( &tmpFrame->fields[2], contents );
#ifdef DEBUG
		printf( "%s TXXX frame already present, overwriting\n", label );
#endif
		return 1;
	}
	else {
		tmpFrame = makeID3Frame( "TXXX", label, contents, NULL );
		id3_tag_attachframe( fileTag, tmpFrame );
#ifdef DEEPDEBUG
		fprintf( stderr, "Attached TXXX frame, nframes is now: %u\n", fileTag->nframes );
#endif
		return 0;
	}
}


int deleteTXXX( struct id3_tag *fileTag, const char *label )
{
	struct id3_frame *tmpFrame;
	
	tmpFrame = findTXXX( fileTag, label );
	if( tmpFrame != NULL ) {
		id3_tag_detachframe( fileTag, tmpFrame );
		id3_frame_delete(tmpFrame);
#ifdef DEBUG
		printf( "Deleted %s TXXX frame\n", label );
#endif
		return 1;
	}
#ifdef DEEPDEBUG
	else {
		printf( "%s TXXX frame not present\n", label );
	}
#endif
	return 0;
}


struct id3_frame * findCOMM( struct id3_tag *fileTag, const char *label )
{
	return findFrameWithLabel( fileTag, "COMM", label, 2 );
}


int setCOMM( struct id3_tag *fileTag, const char *lang, const char *label,
             const char *contents )
{
	struct id3_frame *tmpFrame;
	
	tmpFrame = findCOMM( fileTag, label );
	if( tmpFrame != NULL ) {
		setField( &tmpFrame->fields[1], lang );
		setField( &tmpFrame->fields[3], contents );
#ifdef DEBUG
		printf( "%s COMM frame already present, overwriting\n", label );
#endif
		return 1;
	}
	else {
		tmpFrame = makeID3Frame( "COMM", lang, label, contents );
		id3_tag_attachframe( fileTag, tmpFrame );
#ifdef DEEPDEBUG
		fprintf( stderr, "Attached COMM frame, nframes is now: %u\n", fileTag->nframes );
#endif
		return 0;
	}
}


int deleteCOMM( struct id3_tag *fileTag, const char *label )
{
	struct id3_frame *tmpFrame;
	
	tmpFrame = findCOMM( fileTag, label );
	if( tmpFrame != NULL ) {
		id3_tag_detachframe( fileTag, tmpFrame );
		id3_frame_delete(tmpFrame);
#ifdef DEBUG
		printf( "Deleted %s COMM frame\n", label );
#endif
		return 1;
	}
#ifdef DEEPDEBUG
	else {
		printf( "%s COMM frame not present\n", label );
	}
#endif
	return 0;
}



void tagInfo( const struct id3_tag *fileTag )
{
	float id3v = 0.0;
	switch(id3_tag_version(fileTag)) {
		case 256:
			id3v = 1.1;
			break;
		case 512:
			id3v = 2.2;
			break;
		case 768:
			id3v = 2.3;
			break;
		case 1024:
			id3v = 2.4;
			break;
	}
	printf("Tag version: %1.1f, #frames: %u, paddedsize: %lu\n", id3v, fileTag->nframes, fileTag->paddedsize );	
}


void tagFramesInfo( const struct id3_tag *fileTag )
{
	unsigned int i, j;
	for( i=0; i<fileTag->nframes; i++ ) {
		struct id3_frame *frame = id3_tag_findframe(fileTag, NULL, i);
		fprintf( stderr, "Frame %u: %s with %u fields\n", i, frame->id, frame->nfields );
		if( strcmp(frame->id,"COMM") == 0 ) {
			unsigned char *tmp = getField( id3_frame_field(frame, 2) );
			fprintf( stderr, "  Comment type: %s\n", tmp );
			free(tmp);
		}
		else if( strcmp(frame->id,"TXXX") == 0 ) {
			unsigned char *tmp = getField( id3_frame_field(frame, 1) );
			fprintf( stderr, "  User-defined text type: %s\n", tmp );
			free(tmp);
		}
		fprintf( stderr, "Text encoding: %d\n", id3_field_gettextencoding( &frame->fields[0] ) );
		for( j=0; j<frame->nfields; j++ )
			fprintf( stderr, "Field %u type: %u\n", j, id3_field_type( &frame->fields[j]) );
	}
}


/* Why on earth did they have to use unsigned chars? */
unsigned char * dupString( const unsigned char *str ) {
	if( str == NULL ) {
		fprintf( stderr, "YOU SUCK\n" ); /*DEBUG*/
		return NULL;
	}
	unsigned char *dup = malloc( 1+strlen((char*)str) );
	strncpy( (char*)dup, (char*)str, strlen((char*)str) );
	return dup;
}
